#include "World.h"

#include <algorithm>

#include "Texture2D.h"

// Include our actor class:
#include "StaticMeshActor.h"

// This constructor constructs a world with multiple objects:
World::World()
    : lightCubeMesh(CMAKE_SOURCE_DIR "/models/cube.obj")
    , lightCubeShader(CMAKE_SOURCE_DIR "/shader/lightCube.vert", CMAKE_SOURCE_DIR "/shader/lightCube.frag")
    , runtime(0.f)
{
    // Pre-generate some random light colors:
    for(int i=0; i < 128; ++i){
        float r = (float)rand()/(float)(RAND_MAX);
        float g = (float)rand()/(float)(RAND_MAX);
        float b = (float)rand()/(float)(RAND_MAX);

        // Store light color in lightColors-list:
        lightColors.emplace_back(r,g,b);
    }

    // Override the first light color with a warmer color:
    lightColors[0] = Vec4f(1.0f, 0.9f, 0.8f) * 2.0f;

    // Load container mesh and texture:
    Texture2D containerTexture(CMAKE_SOURCE_DIR "/models/textures/Mat_Base_Color.jpg");
    Mesh containerMesh(CMAKE_SOURCE_DIR "/models/container.obj");

    // Randomly place 200 cube actors into the scene:
    for(int i=0; i < 200; ++i){
        float r1 = (float)rand()/(float)(RAND_MAX);
        float r2 = (float)rand()/(float)(RAND_MAX);
        float r3 = (float)rand()/(float)(RAND_MAX);

        float x = r1 * 30 - 15;
        float y = -0.25f + round(r2*3) * 1.5f;
        float z = r3 * 30 - 15;

        if(abs(x) < 2 && abs(y) < 2 && abs(z) < 2)
            continue;

        Mat4f modelMatrix = Mat4f::translation(x,y,z)* Mat4f::scale(0.5f);

        // Add the cube actor to the actor-list so that it gets rendered:
        actors.emplace_back(StaticMeshActor(containerMesh, containerTexture, modelMatrix));
    }

    // Load ground mesh and texture:
    Texture2D groundTexture(CMAKE_SOURCE_DIR "/models/textures/Concrete_Polished_albedo.jpg");
    Mesh groundMesh(CMAKE_SOURCE_DIR "/models/ground.obj");
    actors.emplace_back(StaticMeshActor(groundMesh, groundTexture, Mat4f::translation(0,-0.225f,0)*Mat4f::scale(3.0)));

    // Load table mesh:
    Mesh tableMesh(CMAKE_SOURCE_DIR "/models/table.obj");
    actors.emplace_back(StaticMeshActor(tableMesh, Mat4f::translation(0,-0.225f,0)));

    // Load island mesh:
    Mesh islandMesh(CMAKE_SOURCE_DIR "/models/island.obj");
    actors.emplace_back(StaticMeshActor(islandMesh, Mat4f::translation(0,-0.225f,0), Vec4f(1.f, 0.83f, 0.f), 200));
}

void World::simulate(){
    runtime += 1.f/60.f;
}


void World::bindLights(Mat4f view, Shader shader, int lightNum){
    // Set the num of lights to be used in the shader:
    shader.setUniform("lightNum", lightNum);

    // Scale the intensity and the range with this factor to make the scene also bright
    // with a few light but not to bright with a lot of lights:
    float intensityfactor = 1 + pow(1 - lightNum / 100.f, 4);

    // Generate light positions and bind them to the shader:
    for(int i=0; i < 100; ++i){
        float t = runtime * 0.1f + i * 75.67f;
        Vec4f lightPosWS(10.f * sin(t * 1.86f), sin(t * 1.38f) * 3.f + 3, 10.f * cos(t * 1.45f));
        Vec4f lightPosCS = view * lightPosWS;

        std::string lightIdxStr = std::to_string(i);
        shader.setUniform("lights["+lightIdxStr+"].position", lightPosCS, 3);
        shader.setUniform("lights["+lightIdxStr+"].color", lightColors[i] * intensityfactor, 3);
        shader.setUniform("lights["+lightIdxStr+"].range", 10.f * intensityfactor);
    }
}

// In this framework, we already assume that the given shader is bound.
void World::render(Mat4f view, Mat4f projection, Shader shader){
    // Now render all actors with the bound lights:
    for(StaticMeshActor& actor : actors){
        actor.render(view, projection, shader);
    }
}

void World::renderLightCubes(Mat4f view, Mat4f projection, int lightNum){
    // Generate light positions and bind them to the shader:
    for(int i=0; i < lightNum; ++i){
        float t = runtime * 0.1f + i * 75.67f;
        Vec4f lightPosWS(10.f * sin(t * 1.86f), sin(t * 1.38f) * 3.f + 3, 10.f * cos(t * 1.45f));

        std::string lightIdxStr = std::to_string(i);
        lightCubeShader.bind();
        lightCubeShader.setUniform("projection", projection);
        lightCubeShader.setUniform("view", view);
        lightCubeShader.setUniform("lightPosition", lightPosWS, 3);
        lightCubeShader.setUniform("lightColor", lightColors[i], 3);
        lightCubeMesh.render();
    }
}
